import java.util.Vector; 
import java.util.List; 
import java.util.StringTokenizer; 
import java.io.*; 

/* Copyright K. Lano 2013-2017
   Package: Class Diagram */ 


public abstract class ModelElement
{ protected String name;
  protected Vector stereotypes = new Vector(); // of String
  protected Vector comments = new Vector(); // of String

  /* attribute kind */
  public static final int NONE = 0; 
  public static final int SEN = 2;
  public static final int INTERNAL = 3;
  public static final int ACT = 7;
  public static final int DERIVED = 5; // no set operations
 
  /* role multiplicity */ 
  public static final int QUALIFIER = -4; 
  public static final int AGGREGATION01 = -3; 
  public static final int AGGREGATION1 = -2; 
  public static final int ZEROONE = -1;
  public static final int MANY = 0;
  public static final int ONE = 1;
 
  /* feature visibility */
  public static final int PUBLIC = 0; 
  public static final int PRIVATE = 1; 
  public static final int PROTECTED = 2; 
  public static final int PACKAGE = 3; 
     

  public ModelElement(String nme)
  { name = nme; }
 
  public String getName()
  { return name; }

  public void setName(String newname)
  { name = newname; }

  public static Vector getNames(Vector elems)
  { Vector res = new Vector(); 
    for (int i = 0; i < elems.size(); i++) 
    { ModelElement e = (ModelElement) elems.get(i); 
      res.add(e.getName()); 
    } 
    return res; 
  }


  public void setStereotypes(Vector s) 
  { stereotypes = s; } 

  public void setStereotypes(String s) 
  { Vector strs = new Vector(); 
    StringTokenizer st = new StringTokenizer(s); 
    while (st.hasMoreTokens())
    { String se = st.nextToken().trim();      
      if (!se.equals("none")) 
      { strs.add(se); }  
    } 
    setStereotypes(strs); 
  } 

  public String getStereotype(int i)
  { if (0 <= i && i < stereotypes.size())
    { return (String) stereotypes.get(i); }
    return null; 
  }

  public Vector getStereotypes()
  { return stereotypes; }

  public void addStereotype(String s)
  { if (stereotypes.contains(s)) { } 
    else 
    { stereotypes.add(s); }
  } 

  public void addStereotypes(Vector ss)
  { for (int i = 0; i < ss.size(); i++) 
    { String s = (String) ss.get(i); 
      addStereotype(s);
    }
  }  

  public void removeStereotype(String s)
  { stereotypes.remove(s); }

  public boolean hasStereotype(String s)
  { return stereotypes.contains(s); }

  public Vector getComments()
  { return comments; }

  public void addComment(String s)
  { comments.add(s); } 

  public void removeComment(String s)
  { comments.remove(s); }

  public boolean isDerived()
  { return stereotypes.contains("derived"); } 

  public String toString()
  { return name; } 

  public static String destring(String s)
  { if (s.length() > 2)
    { if (s.charAt(0) == '"' && s.charAt(s.length()-1) == '"') 
      { return s.substring(1,s.length()-1); } 
    } 
    return s; 
  } 

  public static Type model2type(String val) 
  { String sval = val; 
    if ("void".equals(val)) { return null; } 
    else if ("Integer".equals(val)) { sval = "int"; } 
    else if ("Real".equals(val)) { sval = "double"; } 
    else if ("Boolean".equals(val)) { sval = "boolean"; } 
    else if ("SequenceType".equals(val)) { sval = "Sequence"; } 
    else if ("SetType".equals(val)) { sval = "Set"; } 

    Type typ = new Type(sval, null);
    return typ; 
  }                   

  public void asTextModel(PrintWriter out) { } 

  public boolean equals(Object e) 
  { if (e != null && e instanceof ModelElement)
    { return e.toString().equals(toString()); } 
    return false; 
  } 

  public String saveData()
  { return name; }  // default implementation

  public void saveData(PrintWriter out)
  { out.println(name); }  // default implementation

  public static String baseName(String ename)
  { int ind = ename.indexOf("$");
    if (ind < 0) { return ename; }
    return ename.substring(ind + 1,ename.length());
  }

  public static ModelElement lookupByName(String nme, List mes)
  { for (int i = 0; i < mes.size(); i++) 
    { ModelElement me = (ModelElement) mes.get(i); 
      if (me.getName().equals(nme))
      { return me; } 
    } 
    return null; 
  } 

  public static ModelElement lookupByName(String nme, Vector mes)
  { for (int i = 0; i < mes.size(); i++) 
    { ModelElement me = (ModelElement) mes.get(i); 
      if (me.getName().equals(nme))
      { return me; } 
    } 
    return null; 
  } 

  public static boolean containsName(String nme, Vector mes)
  { ModelElement me = lookupByName(nme,mes); 
    return me != null; 
  } 

  public static String tab(int indent)
  { String res = ""; 
    for (int i = 0; i < indent; i++) 
    { res = res + "  "; } 
    return res; 
  } 

  public abstract void generateJava(PrintWriter out);

  public static int convertCard(String s)
  { int n = 0;
    if ("*".equals(s))
    { return MANY; }
    if ("aggregation 1".equals(s))
    { return AGGREGATION1; } 
    if ("aggregation 0..1".equals(s))
    { return AGGREGATION01; } 
    if ("qualifier".equals(s))
    { return QUALIFIER; } 
    if (s.indexOf(".") < 0)  // not a range
    { try
      { n = Integer.parseInt(s); }
      catch (Exception e)
      { System.err.println("not a valid number");
        return MANY;
      }
      return n;  // ONE is 1
    }
    else
    { return ZEROONE; } // assume
  }

  public static String convertCard(int c)
  { if (c == ZEROONE)
    { return "0..1"; }
    if (c == ONE)
    { return "1"; }
    if (c == MANY)
    { return "*"; }
    if (c == AGGREGATION1)
    { return "aggregation 1"; } 
    if (c == AGGREGATION01)
    { return "aggregation 0..1"; } 
    if (c == QUALIFIER)
    { return "qualifier"; } 
    return ""+c;
  }

  public static int maxCard(int c, int q)
  { if (c == ZEROONE)
    { return 1; } 
    if (c == ONE) 
    { return 1; } 
    if (c == MANY)
    { return q; } 
    return q; // the card of the target entity 
  } 

  public static int mergeMultiplicities(int m1, int m2)
  { // returns closest generalisation of m1, m2: ONE <= ZEROONE <= MANY
    if (m1 == MANY || m2 == MANY) 
    { return MANY; } 
    if (m1 == ZEROONE || m2 == ZEROONE) 
    { return ZEROONE; } 
    return ONE; 
  } 

  public static String capitalise(String nme)
  { String fl = nme.substring(0,1);
    String rem = 
        nme.substring(1,nme.length());
    return fl.toUpperCase() + rem;
  }

  public String constructorParameterCPP()
  { return null; } 
}
